/*

    encapsulate.c, part of
    netpipes: network pipe utilities
    Copyright (C) 1996 Robert Forsman

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */

static char info[] = "encapsulate: a  utility for sockets\nWritten 1996 by Robert Forsman <thoth@purplefrog.com>\n";

/*
 * $Log: encapsulate2.c,v $
 * Revision 1.3  1997/05/15 15:51:24  thoth
 * Switch from longs to shorts for packet lengths.
 * Add an initial protocol negotiation phase (with magic numbers).
 * Add code to prevent packet misalignment.  Ick!
 *
 * Revision 1.2  1997/01/21 21:13:23  thoth
 * Improve some diagnostics.
 * Fix some greivous errors caused by a missing memmove.
 * Add a little more debugging.
 * Make the local return code actually valid.
 * Exit statuses should be obeyed even if the local has
 * no subprocess (the remote might).
 *
 * Revision 1.1  1997/01/15 18:52:34  thoth
 * Initial revision
 *
 */

/*
   Encapsulate

   encapsulate is designed to be used in a script spawned as the child
   of a faucet or hose.

   encapsulate is a program to encapsulate a two-way conversation
   within a two-way socket connection.  The data arriving on the stdin
   of the encapsulate program is encapsulated in a trivial protocol
   and sent to the remote end to be decoded by another encapsulate
   program.  At the same time, the remote encapsulate program is
   encapsulating its stding and sending it over the socket to the
   local program, which extracts it and sends it to stdout.  There are
   special protocol elements to indicate the end of the conversation
   without closing the socket.

   Encapsulate can also be used to connect a subprocess to a socket,
   and, more importantly, transmit the exit status of the subprocess
   to the remote encapsulate program.  This allows one end of a socket
   connection to sample the exit status of a remote process and
   distinguish between "end of data stream" and "program crash/error".

   sample usage

   $0 -fd 3 < /dev/null > outfile

	retrieves outfile from the remote end of socket 3

   $0 -fd 3 > /dev/null < infile

	sends infile to the remote end of socket 3

   $0 -subproc -fd 3 -infd 0 sh -c "sort | uniq -c"

	stdin is sent to the other end of socket 3, and whatever comes
	back is copied to the stdin of the subprocess sh -c "sort |
	uniq -c".  The stdout of the subprocess is the same as the
	stdout of encapsulate.  The exit status of the subprocess is
	transmitted to the remote encapsulate program.

   $0 -fd 3 -outfd 1 tar cvf - /

	the stdout of the tar process is sent to the remote end, and
	its exit status is also transmitted (allowing the remote end
	to know if the backups really completed successfully).


  -remoteonly , -localonly, for exit status stuff.

  Warning.  The htonls are non-portable.  Some of that data could be
  unaligned.  The i386 doesn't care, but a sparc would.  Gotta fix it.

  What if the in&out descriptors for the subproc are the same?  check
  with SSL-auth too.

 */

#include <stdio.h>
#include	<errno.h>
extern int errno;		/* I hate the errno header file */
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "common.h"
#include "memmove.h"

#define EXITCODE_ARGS		127
#define EXITCODE_SYSCALLFAILED	126
#define EXITCODE_EXECFAILED	125
#define EXITCODE_SIGNAL		124
#define EXITCODE_PROTOCOL	123

#define PACKETCODE_EOF	0
#define PACKETCODE_EOF_WAITING	1
#define PACKETCODE_EXITSTATUS	2

typedef	short Nshort;

static int subproc=0;
static int verbose=0;

/********************************************************************/

static void usage()
{
    fprintf(stderr,"Usage : %s --fd n [ -v ] [ --subproc [ --infd n ] [ --outfd n ] [ --preferlocal ] [ --preferremote ] [ --localonly ] [ --remoteonly ] -[#n][v][s[in][on][l][r][L][R]] <command> <args>  ]\n",
	    progname);
}

/********************************************************************/

static int remote_return_code = 0;
static int local_return_code = 0;

static int	child_unreaped = 0;
static int	child_running = 0;
static void probe_child();

#define	MY_BUF_SIZE	4096

int	bufsize=MY_BUF_SIZE;

char	magic_number[4] = {'N', 'p', 'E', 'n' };

/* tell the remote end the maximum packet size (inclusive of length info)
   we're willing to accept.

   We then discover what size it's willing to accept */
int initial_protocol_negotiation(in_fd, out_fd)
     int	in_fd, out_fd;
{
    /* if the remote end won't tell us its size,
       it is required to handle this big at least */
    int	rval=MY_BUF_SIZE;
    int	x;
    unsigned char	buf[sizeof(magic_number) +2];

    if (out_fd>=0) {
	/* tell the remote end how big a packet we'll accept*/
	memcpy(buf, magic_number, sizeof(magic_number));
	buf[4] = (bufsize+sizeof(Nshort))>>8;
	buf[5] = (bufsize+sizeof(Nshort))&0xff;
#ifdef DEBUG
	fprintf(stderr, "%s: negotiating bufsize %d incoming [%02x%02x%02x%02x%02x%02x]\n",
		progname, bufsize,
		buf[0], buf[1], buf[2],
		buf[3], buf[4], buf[5]);
#endif /*DEBUG*/

	x=0;
	while (x<sizeof(buf)) {
	    int	y;
	    y = write(out_fd, buf+x, sizeof(buf) - x);
	    if (y<1) {
		fprintf(stderr, "%s: error during bufsize negotiation [%d=write].  Aborting.  ",
			progname, y);
		perror("");
		exit(EXITCODE_SYSCALLFAILED);
	    }
	    x += y;
	}
    }

    if (in_fd>=0) {
	/* learn how big a packet the remote end is willing to accept */
	x=0;
	while (x<sizeof(buf)) {
	    int	y;
	    y = read(in_fd, buf+x, sizeof(buf) - x);
	    if (y<1) {
		fprintf(stderr, "%s: error during bufsize negotiation [%d=write].  Aborting.  ",
			progname, y);
		if (y!=0)
		    perror("");
		exit(EXITCODE_SYSCALLFAILED);
	    }
	    x += y;
	}

	if (0!= memcmp(buf, magic_number, sizeof(magic_number))) {
	    fprintf(stderr, "%s: magic number mismatch during bufsize negotiation.  Aborting.  ",
		    progname);
	    exit(EXITCODE_PROTOCOL);
	}

	rval = buf[sizeof(magic_number)]<<8 |
	    buf[sizeof(magic_number)+1];
	rval -= sizeof(Nshort);		/* subtract off the length field */

#ifdef DEBUG
	fprintf(stderr, "%s: negotiated bufsize %d incoming [%02x%02x%02x%02x%02x%02x]\n",
		progname, rval,
		buf[0], buf[1], buf[2],
		buf[3], buf[4], buf[5]);
#endif /*DEBUG*/
    }

    return rval;
}

static void main_io_loop( in_fd, sock_fd, out_fd)
{
    /* bytes flow from the buffer listed first to the last one */
    char	*buf_from_local;
    char	*buf_to_remote;
    /* remote end is between these two buffers */
    char	*buf_from_remote;
    char	*buf_to_local;

    int	bfl_len=0;
    int	btr_len=0;
    int	bfr_len=0;
    int	btl_len=0;

    int sock_closed=0;
    fd_set	readfds, writefds;
    int	maxfd;

    int	out_bufsize = initial_protocol_negotiation
	((out_fd<0) ? -1 : sock_fd,
	 ( in_fd<0) ? -1 : sock_fd );

    buf_from_local = malloc(out_bufsize);
    buf_to_remote = malloc(out_bufsize+sizeof(Nshort));
    buf_from_remote = malloc(bufsize+sizeof(Nshort));
    buf_to_local = malloc(bufsize);

    if (in_fd < 0)		/* don't read stdin */
      in_fd = -3;		/* don't even attempt to write to the socket */

    while (1) {
	int	rval;

	FD_ZERO(&readfds);
	FD_ZERO(&writefds);
	maxfd = -1;

	if (in_fd<0 ||	/* stdin is closed, nothing to read */
	    bfl_len>= out_bufsize/* the incoming buffer is full */
	    ) {
	    /* don't bother adding STDIN */
	} else {
	    FD_SET(in_fd, &readfds);
	    /* we know maxfd<0 */
	    maxfd = in_fd;
	}

	if (btr_len != 0) {	/* we have bytes to write to the remote end */
	    FD_SET(sock_fd, &writefds);
	    if (sock_fd>maxfd) maxfd=sock_fd;
	}

	if (sock_closed ||
	    bfr_len >= bufsize+sizeof(Nshort)) {
	    /* don't bother adding the read socket */
	} else {
	    FD_SET(sock_fd, &readfds);
	    if (sock_fd>maxfd) maxfd=sock_fd;
	}

	if (btl_len != 0) {	/* we have bytes to write to the local end */
	    FD_SET(out_fd, &writefds);
	    if (out_fd>maxfd) maxfd=out_fd;
	}

	if (maxfd<0
	    && in_fd == -3	/* make sure we've consummated the
				   conversation and waited on the child. */
	    ) {
	    if (bfl_len != 0 ||
		btr_len != 0 ||
		bfr_len != 0 ||
		btl_len != 0) {
		fprintf(stderr, "%s: leftover bytes in buffers at end of encapsulation.  That is bad because it means Bob made a logic error in his code.\n", progname);
	    }
	    break;		/* run out of things to do */
	}

	/* I think there are still race conditions here */
	if ( subproc && in_fd == -2 && !child_running
	     && child_unreaped) {
	    probe_child();
	    FD_ZERO(&readfds);
	    FD_ZERO(&writefds);
	    rval = 1;		/* fake it. */
	} else if (maxfd>0) {
	    rval = select(maxfd+1, &readfds, &writefds, (fd_set*)0,
			  (struct timeval*)0);
	}

	if (rval<0) {
	    if (errno == EINTR) {
		continue;
	    } else {
		fprintf(stderr, "%s: error during select: ", progname);
		perror("");
		exit(EXITCODE_SYSCALLFAILED);
	    }
	    /* NOTREACHED */
	} else if (rval==0) {
	    /* out of file descriptors, clearly */
	    return;
	}

	/* read bytes from stdin */
	if (FD_ISSET(in_fd, &readfds)) {
#ifdef DEBUG
	    fprintf(stderr, "%s: read(in_fd, buf_from_local+%d, %d-%d)",
			progname, bfl_len, out_bufsize, bfl_len);
#endif
	    rval = read(in_fd, buf_from_local+bfl_len, out_bufsize-bfl_len);
#ifdef DEBUG
		fprintf(stderr, "=%d;\n", rval);
#endif
	    if (rval==0) {
#ifdef DEBUG
		fprintf(stderr, "%s: `closing' stdin.\n", progname);
#endif
		in_fd = -1;
	    } else if (rval<0) {
		if (errno!=EINTR) {
		    fprintf(stderr, "%s: error during read(0,,).  Aborting.  ", progname );
		    perror("");
		    exit(EXITCODE_SYSCALLFAILED);
		}
		/* never mind. try again later */
	    } else {
		bfl_len += rval;
	    }
	}

	/* write bytes to the socket */
	if (FD_ISSET(sock_fd, &writefds)) {
#ifdef DEBUG
	    fprintf(stderr, "%s: write(sock_fd, buf_to_remote, %d)",
		    progname, btr_len);
#endif
	    rval = write(sock_fd, buf_to_remote, btr_len);
#ifdef DEBUG
	    fprintf(stderr, "=%d;\n", rval);
#endif
	    if (rval==0) {
		fprintf(stderr, "%s: Inexplicable!  write(sock_fd,,%d) returns 0\n",
			progname, btr_len);
		exit(EXITCODE_SYSCALLFAILED);
	    } else if (rval<0) {
		if (errno!=EINTR) {
		    fprintf(stderr, "%s: error during write(sock_fd,,%d).  Aborting.  ",
			    progname, btr_len);
		    perror("");
		    exit(EXITCODE_SYSCALLFAILED);
		}
		/* never mind. try again later */
	    } else {
		btr_len -= rval;
		memmove(buf_to_remote, buf_to_remote+rval, btr_len);
	    }
	}

	/* read bytes from the socket */
	if (FD_ISSET(sock_fd, &readfds)) {
	    int	desired_size =0;

	    /* the following big scary mess of a block is to prevent
               us from reading too many bytes from the socket which
               would corrupt later encapsulations.

	       I believe that this code will never ask the read to get
	       more bytes than are necessary to get a complete EOF or
	       EXITCODE packet.  As long as the packet is in the
	       queue, the code will choose exactly as many bytes are
	       necessary to complete the packet.  If the packet is
	       removed from the queue, the code that did it should
	       have set sock_closed as a matter of course, and we
	       won't be reading from the socket after that.

	       */

	    {
		int	offset = 0;
#ifdef DEBUG
		fprintf(stderr, "%s: Scanning:", progname);
#endif
		while (offset<bfr_len) {
		    /* this could really suck if the offset
		       crossed an alignment buffer.
		       prevent that by making all packets a multiple
		       of sizeof(Nshort) */
		    Nshort *	pkt = (Nshort*)(offset+buf_from_remote);
#ifdef DEBUG
		    fprintf(stderr, " %d", offset);
#endif

		    if (offset+sizeof(Nshort) > bfr_len) {
			desired_size = offset+sizeof(Nshort)+1; /* smallest packet is 3 bytes */
			break;
		    } else {
			int	pktlen = ntohs(*pkt);
			if (pktlen==0) {
			    /* special packet */
			    if (offset+2*sizeof(Nshort) >bfr_len) {
				desired_size = offset+2*sizeof(Nshort);
			    } else {
				switch (ntohs(pkt[1])) {
				case PACKETCODE_EOF:
				    desired_size = offset+2*sizeof(Nshort);
				    break;
				case PACKETCODE_EOF_WAITING:
				    desired_size = offset
					+ 2*sizeof(Nshort) /* this packet */
					+ 3*sizeof(Nshort) /* the impending P*C*_EXITSTATUS*/
					;
				    break;
				case PACKETCODE_EXITSTATUS:
				    desired_size = offset+3*sizeof(Nshort);
				    break;
				default:
				    fprintf(stderr, "%s: encapsulation protocol error.  Unknown special packet 0x%lx\n",
					    progname, (long)ntohs(pkt[1]));
				    exit(EXITCODE_PROTOCOL);
				}
			    }
			    break;
			} else {
			    /* normal packet */
			    offset += sizeof(Nshort)+pktlen;
			    desired_size = offset+sizeof(Nshort); /* snag next length packet */
			}
		    }
		}
#ifdef DEBUG
		fprintf(stderr, " %d,%d ", offset, desired_size);
#endif
		if (desired_size ==0)
		    desired_size = offset;
		if (desired_size ==0)
		    desired_size = sizeof(Nshort)+1; /* smallest packet is 5 bytes */
		if (desired_size > bufsize+sizeof(Nshort))
		    desired_size = bufsize+sizeof(Nshort);
#ifdef DEBUG
		fprintf(stderr, "%d\n", desired_size);
#endif
	    }

	    /* now we know how many bytes we can read without corrupting a later conversation */

	    if (desired_size>bfr_len) {
#ifdef DEBUG
		fprintf(stderr, "%s: read(sock_fd, buf_from_remote+%d, %d-%d)",
			progname, bfr_len, desired_size, bfr_len);
#endif
		rval = read(sock_fd, buf_from_remote+bfr_len, desired_size-bfr_len);
#ifdef DEBUG
		fprintf(stderr, "=%d;\n", rval);
#endif
		if (rval==0) {
		    fprintf(stderr, "%s: encapsulation protocol error reading socket.  Socket closed without an EOF packet.\n", progname);
		    sock_closed = 1;
		    exit(EXITCODE_PROTOCOL);
		} else if (rval<0) {
		    if (errno!=EINTR) {
			fprintf(stderr, "%s: error during read(sock_fd,,%d).  Aborting.  ",
				progname, desired_size-bfr_len);
			perror("");
			exit(EXITCODE_SYSCALLFAILED);
		    }
		    /* never mind. try again later */
		} else {
		    bfr_len += rval;
		}
	    }
	}

	/* write bytes to stdout */
	if (FD_ISSET(out_fd, &writefds)) {
#ifdef DEBUG
	    fprintf(stderr, "%s: write(out_fd, buf_to_local, %d)",
		    progname, btl_len);
#endif
	    rval = write(out_fd, buf_to_local, btl_len);
#ifdef DEBUG
	    fprintf(stderr, "=%d;\n", rval);
#endif
	    if (rval==0) {
		fprintf(stderr, "%s: Inexplicable!  write(out_fd,,%d) returns 0\n",
			progname, btl_len);
		exit(EXITCODE_SYSCALLFAILED);
	    } else if (rval<0) {
		if (errno!=EINTR) {
		    fprintf(stderr, "%s: error during write(out_fd,,%d).  Aborting.  ",
			    progname, btl_len);
		    perror("");
		    exit(EXITCODE_SYSCALLFAILED);
		}
		/* never mind. try again later */
	    } else {
		btl_len -= rval;
		memmove(buf_to_local, buf_to_local+rval, btl_len);
	    }
	}

	/* encapsulate */
	if (btr_len == 0) {
	    /* we can encapsulate a new packet */
	    if (bfl_len>0) {
		/* prevent misalignments of packets in
		   memory on remote machine */
		int pktlen = bfl_len;
		if (in_fd>=0)
		    pktlen -= bfl_len%sizeof(Nshort);

		if (pktlen) {
		  *((Nshort*)buf_to_remote) = htons(pktlen);
		  memcpy(buf_to_remote+sizeof(Nshort), buf_from_local, pktlen);
		  btr_len = pktlen+sizeof(Nshort);
		  if (pktlen<bfl_len)
		    memmove(buf_from_local, buf_from_local+pktlen, bfl_len - pktlen);
		  bfl_len -= pktlen;
		}
	    } else if ( in_fd == -1 || in_fd == -2 ) {
		Nshort *	pkt = (Nshort*)buf_to_remote;
		/* send the special termination packet */
		pkt[0] = htons(0);
		if (subproc) {
		    if (!child_unreaped) {
			pkt[1] = htons(PACKETCODE_EXITSTATUS);
			pkt[2] = htons(local_return_code);
			btr_len = sizeof(Nshort)* 3;
			in_fd = -3; /* we've finally closed the connection */
#ifdef DEBUG
			fprintf(stderr, "%s: `closing' stdin[2 ES = %d].\n",
				progname, local_return_code);
#endif
		    } else if (in_fd == -1) {
			pkt[1] = htons(PACKETCODE_EOF_WAITING);
			btr_len = sizeof(Nshort)* 2;
			in_fd = -2; /* we've almost closed the connection */
#ifdef DEBUG
			fprintf(stderr, "%s: `closing' stdin[2 EW].\n",
				progname);
#endif
		    }
		} else {
		    pkt[1] = htonl(PACKETCODE_EOF);
		    btr_len = sizeof(Nshort)* 2;
		    in_fd = -3;	/* we've now closed the connection */
#ifdef DEBUG
		    fprintf(stderr, "%s: `closing' stdin[2 EOF].\n", progname);
#endif
		}
	    }
	}
	/* decapsulate */
	if (btl_len == 0 && bfr_len>sizeof(Nshort)) {
	    btl_len = ntohs(*((Nshort*)buf_from_remote));
	    if (btl_len==0) {
		Nshort * pkt = (Nshort*)buf_from_remote;
		if (bfr_len >=2*sizeof(Nshort) ) {
		    switch (ntohs(pkt[1])) {
		    case PACKETCODE_EOF:
#ifdef DEBUG
			fprintf(stderr, "%s: `closing' socket[2].\n", progname);
#endif
			bfr_len -= 2*sizeof(Nshort);
			memmove(buf_from_remote, buf_from_remote+2*sizeof(Nshort), bfr_len);
			close(out_fd); /* we can do this */
			sock_closed = 1;
			break;
		    case PACKETCODE_EOF_WAITING:
#ifdef DEBUG
			fprintf(stderr, "%s: `closing' socket[2 EW].\n", progname);
#endif
			bfr_len -= 2*sizeof(Nshort);
			memmove(buf_from_remote, buf_from_remote+2*sizeof(Nshort), bfr_len);
			close(out_fd); /* we can do this */

			/* can't do this yet.  Gotta wait for the
                           PACKETCODE_EXITSTATUS.

			   sock_closed = 1; */

			break;
		    case PACKETCODE_EXITSTATUS:
			if (bfr_len>=3*sizeof(Nshort)) {
			    remote_return_code =  ntohs(pkt[2]);
			    bfr_len -= 3*sizeof(Nshort);
			    memmove(buf_from_remote, buf_from_remote+3*sizeof(Nshort), bfr_len);
#ifdef DEBUG
			    fprintf(stderr, "%s: got exitstatus of %d.\n",
				    progname, remote_return_code);
#endif
			    close(out_fd); /* we can do this */
			    sock_closed = 1;
			}
			break;
		    default:
			fprintf(stderr, "%s: encapsulation protocol error.  Unknown special packet 0x%lx\n",
				progname, (long)ntohs(pkt[1]));
			exit(EXITCODE_PROTOCOL);
		    }
		}
	    } else if (bfr_len >= sizeof(Nshort)+btl_len) {
		memcpy(buf_to_local, buf_from_remote+sizeof(Nshort), btl_len);
		bfr_len -= sizeof(Nshort)+btl_len;
		memmove(buf_from_remote, buf_from_remote+btl_len+sizeof(Nshort), bfr_len);
	    } else {
		/* not enough bytes to have a full packet */
		btl_len = 0;
	    }
	}
    }
}

/********************************************************************/

static int	childpid = -1;

static void waitonchild()
{
    /* got a SIGCHILD.
       It must be that: */
    child_running = 0;
}

static void probe_child()
{
    if (child_running || !child_unreaped)
	return;

    if ( 0>=wait(&local_return_code)) {
	fprintf(stderr, "%s: wait returned error or zero: ", progname);
	perror("");
	exit(EXITCODE_SYSCALLFAILED);
    }
    if (!WIFEXITED(local_return_code))
	local_return_code = EXITCODE_SIGNAL;
    else
	local_return_code = WEXITSTATUS(local_return_code);
    child_unreaped = 0;
}

static void spawn_child(sp_infd, sp_outfd, parent_inp, parent_outp, cmd)
     int	sp_infd, sp_outfd;
     int	*parent_inp, *parent_outp;
     char	**cmd;
{
    int	tochild[2];
    int	fromchild[2];

    signal(SIGCHLD,waitonchild);
    child_running = 1;		/* well, not yet. */
    child_unreaped = 1;

    if (sp_infd>=0) {
	if (pipe(tochild) !=0) {
	    fprintf(stderr, "%s: totally failed to pipe(2): ", progname);
	    perror("");
	    exit (EXITCODE_SYSCALLFAILED);
	}
	*parent_outp = tochild[1];
    }
    if (sp_outfd>=0) {
	if (pipe(fromchild) !=0) {
	    fprintf(stderr, "%s: totally failed to pipe(2): ", progname);
	    perror("");
	    exit (EXITCODE_SYSCALLFAILED);
	}
	*parent_inp = fromchild[0];
    }

    childpid = fork();
    if (childpid<0) {
	fprintf(stderr, "%s: unable to fork: ", progname);
	perror("");
	/* I would clear child_running, but, look at the next line */
	exit(EXITCODE_SYSCALLFAILED);
    }

    /* now there's a child running (assuming no race conditions, which
       is why I set it up above and not here.  I'm stupid, but
       paranoid). */

    if (childpid==0) {
	/* child */
	if (sp_infd>=0) {
	    close(tochild[1]);
	    dup2(tochild[0], sp_infd);
	    close(tochild[0]);
	}
	if (sp_outfd>=0) {
	    close(fromchild[0]);
	    dup2(fromchild[1], sp_outfd);
	    close(fromchild[1]);
	}

	execvp(*cmd, cmd);
	fprintf(stderr, "%s: Unable to exec %s: ", progname, *cmd);
	perror("");

	exit(EXITCODE_EXECFAILED);

    } else {
	/* parent */

	if (sp_infd>=0) {
	    close(tochild[0]);
	    close(1);
	}
	if (sp_outfd>=0) {
	    close(fromchild[1]);
	    close(0);
	}
    }
}

/********************************************************************/

static int scan_flag_numeric_fd(s, fdp)
    char *s;
    int	*fdp;
{
    int	n;
    if (1 != sscanf(s, "%i%n", fdp, &n)) {
	fprintf(stderr, "%s: parse error in file descriptor list at '%s'\n", progname, s);
	exit(EXITCODE_ARGS);
    }
    return n;
}

/********************************************************************/

enum mergereturns_ {
    UNINITIALIZED,
    PREFER_LOCAL,		/* if both local and remote processes
				   error, return the local code */
    PREFER_REMOTE,		/* if both local and remote processes
				   error, return the remote code.  */
    LOCAL_ONLY,			/* return the exit code of the local
				   process, ignoring the return code
				   of the remote process. */
    REMOTE_ONLY,		/* return the exit code of the remote
				   process, ignoring the return code
				   of the local process. */
} merge_returns = UNINITIALIZED;

static int	sockfd = -1;
static int	infd= -1, outfd= -1;

int main (argc,argv)
     int argc;
     char ** argv;
     
{
    set_progname(argv[0]);

    if (sizeof(Nshort) != 2) {
	fprintf(stderr, "%s: greivous porting error. sizeof(Nshort) == %d, not 2.\n",
		progname, sizeof(Nshort));
	exit(EXITCODE_ARGS);
    }

    while (argc>1) {
	char *arg = argv[1];
	if (arg[0] != '-') {
	    break;
	}
	arg++;
	if (arg[0] == '-') {
	    arg++;
	    if (0==strcmp(arg, "verbose")) {
		verbose = 1;
		argv++; argc--;
	    } else if (0==strcmp(arg, "fd")) {
		argv++; argc--;
		if (argc<2) {
		    fprintf(stderr, "%s: --fd requires file number for socket.\n",
			    progname);
		    usage();
		    exit(EXITCODE_ARGS);
		} else {
		    sockfd = atoi(argv[1]);
		    argv++; argc--;
		}
	    } else {
		/* unknown -- flag.  Assume it's a command :) */
		break;
	    }
	} else {
	    /* it's a set of single dash flags. */
	    do { switch (arg[0]) {
	    case '#':
		arg += scan_flag_numeric_fd(arg+1, &sockfd);
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 's':
		subproc=1;
		break;
	    case 'i':
		arg += scan_flag_numeric_fd(arg+1, &infd);
		break;
	    case 'o':
		arg += scan_flag_numeric_fd(arg+1, &outfd);
		break;
	    case 'l':
		merge_returns = PREFER_LOCAL;
		break;
	    case 'r':
		merge_returns = PREFER_REMOTE;
		break;
	    case 'L':
		merge_returns = LOCAL_ONLY;
		break;
	    case 'R':
		merge_returns = REMOTE_ONLY;
		break;
	    case 0:
		fprintf(stderr, "%s: blank compact flag.\n", progname);
		/* fall through */
	    default:
		fprintf(stderr, "%s: Unknown compact flag beginning %s\n", progname, arg);
		usage();
		exit (EXITCODE_ARGS);
	    } arg++;
	    } while (arg[0]);

	    argv++;
	    argc--;
	}
    }

    /* argv+1 points to an unrecognized flag that must be the
       subprocess cmd and arguments. */

    if (argc>1 && !subproc) {
	fprintf(stderr, "%s: Unknown flag %s\n", progname, argv[1]);
	usage();
	exit (EXITCODE_ARGS);
    }

    if (sockfd<0) {
	fprintf(stderr, "%s:  I must know the file number for the socket.\n",
		progname);
	usage();
	exit(EXITCODE_ARGS);
    }

    if (subproc) {
	if (merge_returns == UNINITIALIZED)
	    merge_returns = PREFER_LOCAL;
	/* check to make sure at least one descriptor is rerouted */
	if (infd<0 && outfd<0) {
	    fprintf(stderr, "%s: must redirect at least one descriptor of subprocess.  (-infd or -outfd)\n", progname);
	    usage();
	    exit(EXITCODE_ARGS);
	}
    } else {
	if (merge_returns == PREFER_LOCAL ||
	    merge_returns == PREFER_REMOTE ||
	    merge_returns == LOCAL_ONLY) {
	    fprintf(stderr, "%s: no local process to get a return code from\n", progname);
	    usage();
	    exit (EXITCODE_ARGS);
	} else if (merge_returns == UNINITIALIZED) {
	    merge_returns = REMOTE_ONLY;
	}
    }

    if (verbose)
        emit_version("encapsulate", 1996);

    {
	int	my_in = 0;
	int	my_out = 1;

	if (subproc)
	    spawn_child(infd, outfd, &my_in, &my_out, argv+1);

	main_io_loop(my_in, sockfd, my_out);
    }

    while (child_running) {
	pause();
    }
    probe_child();

    switch (merge_returns) {
    case PREFER_LOCAL:
	if (local_return_code!=0)
	    exit (local_return_code);
	else
	    exit (remote_return_code);
	/* NOTREACHED */
    case PREFER_REMOTE:
	if (remote_return_code!=0)
	    exit (remote_return_code);
	else
	    exit (local_return_code);
	/* NOTREACHED */
    case LOCAL_ONLY:
	exit(local_return_code);
    case REMOTE_ONLY:
	exit(remote_return_code);
    default:
	fprintf(stderr, "%s: logic error.  merge_returns has bogus value.\n",
		progname);
	exit(EXITCODE_ARGS);
    }
}
