/* watch.c
   watch [-012] pid-file program [ arg ... ]

   $Id: watch.c,v 1.4 1997/02/12 15:28:22 hanche Exp $

   If pid-file exists, watch exits immediately.  Otherwise, it puts
   itself in the background and dissociates itself from the
   controlling tty, and runs program with the specified args,
   recording the pid of program in the pid-file.  When the program
   exits, watch removes the pid-file.

   If the pid-file equals '' (the empty string then no pid-file is opened,
   and watch does not wait for the program.

   The flag -0 causes watch not to close stdin.
   The flag -1 causes watch not to close stdout.
   The flag -2 causes watch not to close stderr.
   The default is to close the standard file descriptors
   and reopen them on /dev/null.
   Do not use these flags with the corresponding file descriptor
   open on the controlling tty, or tty dissociation may fail.

   This program avoids stdio in an attempt to be tiny.

   TO COMPILE:
   ===========
   gcc -DHASSIGACTION -o watch watch.c
   If that fails, try without -DHASSIGACTION.  (Or with cc replacing gcc).

   $Log: watch.c,v $
   Revision 1.4  1997/02/12 15:28:22  hanche
   Added the option of not keeping a pid-file and not waiting.
   This is useful for, e.g., putting a logger program in the background.

   Revision 1.3  1997/02/04 22:54:53  hanche
   Stupid bug fixed.  Stupid, stupid!  (Inappropriate reuse of fd).

   Revision 1.2  1996/12/13 15:51:16  hanche
   Fixed a possibility of a race condition.  Apparently, on HP-UX 10.X
   each rc script is run in a pty which is closed when the script
   finishes.  Thus, the child process might receive a SIGHUP before it
   has the time to dissociate from the controlling tty.  To avoid this,
   the parent waits for the child to send a SIGUSR1 before exiting.
   Thanks Bert Gijsbers for pointing out the problem.

   I also added the much asked for feature of allowing some of the
   standard file descriptors not to be closed, for those cases when the
   user wishes to redirect these to a file.

   Revision 1.1  1996/12/13 14:24:22  hanche
   Initial revision

*/

#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>

static char *usage="Usage: watch [-012] pid-file program [ arg ... ]\n";

static void
exit_msg(int status, char *msg) { write(2,msg,strlen(msg)); exit(status); }

/* Lifted from qmail's signal.c */
static void
catch(int sig, void (*f)())
{
#ifdef HASSIGACTION
  struct sigaction sa;
  sa.sa_handler = f;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  sigaction(sig,&sa,(struct sigaction *) 0);
#else
  signal(sig,f); /* won't work under System V, even nowadays---dorks */
#endif
}

static void
signal_usr1() { _exit(0); }

static void
putint(int fd, int n) {
  int neg=(n<0), i=20;
  char s[20];
  s[--i]='\n';
  if (neg) n=-n;
  do {s[--i]='0'+n%10; n=n/10;} while(n);
  write(fd, s+i,20-i); }

int
main(int argc, char *argv[]) {
  char *pidfile, *prog, **args, *cp;
  int pidfd, fd, fdnull, child_pid, status, close_fd[3];
  for (fd=0; fd<3; fd++) close_fd[fd]=1;
  for (--argc, ++argv; argc>0 && **argv=='-'; --argc, ++argv) {
    for (cp=(*argv)+1; *cp!=0; cp++) {
      switch (*cp) {
      case '0': close_fd[0]=0; break;
      case '1': close_fd[1]=0; break;
      case '2': close_fd[2]=0; break;
      default: exit_msg(1,usage); }}}
  if (argc<2) exit_msg(1,usage);
  pidfile=argv[0];
  prog=argv[1];
  args=argv+1;
  if (pidfile[0] == 0)
    pidfd=-1;
  else {
    pidfd=open(pidfile,O_WRONLY|O_CREAT|O_EXCL,0644);
    if (pidfd<0) exit_msg(2,"Exclusive open of pid-file failed\n"); }
  catch(SIGUSR1, &signal_usr1);
  switch(fork()) {
  case -1: exit_msg(3,"Fork failed"); break;
  case 0: /* We're the child */
    catch(SIGUSR1, SIG_DFL);
    setsid();
    for (fdnull=-1, fd=0; fd<3; fd++)
      if (close_fd[fd]) {
	close(fd);
	if (fdnull<0) {
	  fdnull=open("/dev/null", O_RDWR, 666);
	  if (fdnull<0) _exit(2); }
	if (fdnull != fd) dup2(fdnull, fd); }
    kill(getppid(),SIGUSR1);
    child_pid=fork();
    switch(child_pid) {
    case -1: _exit(1);
    case 0: /* We're the grandchild */
      if (pidfd > 0) close(pidfd);
      execvp(prog,args);
      _exit(1);
    default: /* We're still the child */
      if (pidfd < 0) _exit(0);
      putint(pidfd, child_pid);
      close(pidfd);
      wait(&status);
      unlink(pidfile); }
  default: /* We're the original parent */
    wait(&status); }
  _exit(0);
}
