#ifndef lint
static char *rcsid =
	"@(#)$Header: traceroute.c,v 1.19 93/04/23 13:41:15 boone Exp $ (PSC)";
#endif

/*
 * traceroute host  - trace the route ip packets follow going to "host".
 *
 * Attempt to trace the route an ip packet would follow to some
 * internet host.  We find out intermediate hops by launching probe
 * packets with a small ttl (time to live) then listening for an
 * icmp "time exceeded" reply from a gateway.  We start our probes
 * with a ttl of one and increase by one until we get an icmp "port
 * unreachable" (which means we got to "host") or hit a max (which
 * defaults to 30 hops & can be changed with the -m flag).  Three
 * probes (change with -q flag) are sent at each ttl setting and a
 * line is printed showing the ttl, address of the gateway and
 * round trip time of each probe.  If the probe answers come from
 * different gateways, the address of each responding system will
 * be printed.  If there is no response within a 5 sec. timeout
 * interval (changed with the -w flag), a "*" is printed for that
 * probe.
 *
 * Probe packets are UDP format.  We don't want the destination
 * host to process them so the destination port is set to an
 * unlikely value (if some clod on the destination is using that
 * value, it can be changed with the -p flag).
 *
 * A sample use might be:
 *
 *     [yak 71]% traceroute nis.nsf.net.
 *     traceroute to nis.nsf.net (35.1.1.48), 30 hops max, 56 byte packet
 *      1  helios.ee.lbl.gov (128.3.112.1)  19 ms  19 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  39 ms  19 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  39 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  39 ms  40 ms  39 ms
 *      5  ccn-nerif22.Berkeley.EDU (128.32.168.22)  39 ms  39 ms  39 ms
 *      6  128.32.197.4 (128.32.197.4)  40 ms  59 ms  59 ms
 *      7  131.119.2.5 (131.119.2.5)  59 ms  59 ms  59 ms
 *      8  129.140.70.13 (129.140.70.13)  99 ms  99 ms  80 ms
 *      9  129.140.71.6 (129.140.71.6)  139 ms  239 ms  319 ms
 *     10  129.140.81.7 (129.140.81.7)  220 ms  199 ms  199 ms
 *     11  nic.merit.edu (35.1.1.48)  239 ms  239 ms  239 ms
 *
 * Note that lines 2 & 3 are the same.  This is due to a buggy
 * kernel on the 2nd hop system -- lbl-csam.arpa -- that forwards
 * packets with a zero ttl.
 *
 * A more interesting example is:
 *
 *     [yak 72]% traceroute allspice.lcs.mit.edu.
 *     traceroute to allspice.lcs.mit.edu (18.26.0.115), 30 hops max
 *      1  helios.ee.lbl.gov (128.3.112.1)  0 ms  0 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  19 ms  19 ms  19 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  19 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  19 ms  39 ms  39 ms
 *      5  ccn-nerif22.Berkeley.EDU (128.32.168.22)  20 ms  39 ms  39 ms
 *      6  128.32.197.4 (128.32.197.4)  59 ms  119 ms  39 ms
 *      7  131.119.2.5 (131.119.2.5)  59 ms  59 ms  39 ms
 *      8  129.140.70.13 (129.140.70.13)  80 ms  79 ms  99 ms
 *      9  129.140.71.6 (129.140.71.6)  139 ms  139 ms  159 ms
 *     10  129.140.81.7 (129.140.81.7)  199 ms  180 ms  300 ms
 *     11  129.140.72.17 (129.140.72.17)  300 ms  239 ms  239 ms
 *     12  * * *
 *     13  128.121.54.72 (128.121.54.72)  259 ms  499 ms  279 ms
 *     14  * * *
 *     15  * * *
 *     16  * * *
 *     17  * * *
 *     18  ALLSPICE.LCS.MIT.EDU (18.26.0.115)  339 ms  279 ms  279 ms
 *
 * (I start to see why I'm having so much trouble with mail to
 * MIT.)  Note that the gateways 12, 14, 15, 16 & 17 hops away
 * either don't send ICMP "time exceeded" messages or send them
 * with a ttl too small to reach us.  14 - 17 are running the
 * MIT C Gateway code that doesn't send "time exceeded"s.  God
 * only knows what's going on with 12.
 *
 * The silent gateway 12 in the above may be the result of a bug in
 * the 4.[23]BSD network code (and its derivatives):  4.x (x <= 3)
 * sends an unreachable message using whatever ttl remains in the
 * original datagram.  Since, for gateways, the remaining ttl is
 * zero, the icmp "time exceeded" is guaranteed to not make it back
 * to us.  The behavior of this bug is slightly more interesting
 * when it appears on the destination system:
 *
 *      1  helios.ee.lbl.gov (128.3.112.1)  0 ms  0 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  19 ms  39 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  19 ms  39 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  39 ms  40 ms  19 ms
 *      5  ccn-nerif35.Berkeley.EDU (128.32.168.35)  39 ms  39 ms  39 ms
 *      6  csgw.Berkeley.EDU (128.32.133.254)  39 ms  59 ms  39 ms
 *      7  * * *
 *      8  * * *
 *      9  * * *
 *     10  * * *
 *     11  * * *
 *     12  * * *
 *     13  rip.Berkeley.EDU (128.32.131.22)  59 ms !  39 ms !  39 ms !
 *
 * Notice that there are 12 "gateways" (13 is the final
 * destination) and exactly the last half of them are "missing".
 * What's really happening is that rip (a Sun-3 running Sun OS3.5)
 * is using the ttl from our arriving datagram as the ttl in its
 * icmp reply.  So, the reply will time out on the return path
 * (with no notice sent to anyone since icmp's aren't sent for
 * icmp's) until we probe with a ttl that's at least twice the path
 * length.  I.e., rip is really only 7 hops away.  A reply that
 * returns with a ttl of 1 is a clue this problem exists.
 * Traceroute prints a "!" after the time if the ttl is <= 1.
 * Since vendors ship a lot of obsolete (DEC's Ultrix, Sun 3.x) or
 * non-standard (HPUX) software, expect to see this problem
 * frequently and/or take care picking the target host of your
 * probes.
 *
 * Other possible annotations after the time are !H, !N, !P (got a host,
 * network or protocol unreachable, respectively), !S or !F (source
 * route failed or fragmentation needed -- neither of these should
 * ever occur and the associated gateway is busted if you see one).  If
 * almost all the probes result in some kind of unreachable, traceroute
 * will give up and exit.
 *
 * Notes
 * -----
 * This program must be run by root or be setuid.  (I suggest that
 * you *don't* make it setuid -- casual use could result in a lot
 * of unnecessary traffic on our poor, congested nets.)
 *
 * This program requires a kernel mod that does not appear in any
 * system available from Berkeley:  A raw ip socket using proto
 * IPPROTO_RAW must interpret the data sent as an ip datagram (as
 * opposed to data to be wrapped in a ip datagram).  See the README
 * file that came with the source to this program for a description
 * of the mods I made to /sys/netinet/raw_ip.c.  Your mileage may
 * vary.  But, again, ANY 4.x (x < 4) BSD KERNEL WILL HAVE TO BE
 * MODIFIED TO RUN THIS PROGRAM.
 *
 * The udp port usage may appear bizarre (well, ok, it is bizarre).
 * The problem is that an icmp message only contains 8 bytes of
 * data from the original datagram.  8 bytes is the size of a udp
 * header so, if we want to associate replies with the original
 * datagram, the necessary information must be encoded into the
 * udp header (the ip id could be used but there's no way to
 * interlock with the kernel's assignment of ip id's and, anyway,
 * it would have taken a lot more kernel hacking to allow this
 * code to set the ip id).  So, to allow two or more users to
 * use traceroute simultaneously, we use this task's pid as the
 * source port (the high bit is set to move the port number out
 * of the "likely" range).  To keep track of which probe is being
 * replied to (so times and/or hop counts don't get confused by a
 * reply that was delayed in transit), we increment the destination
 * port number before each probe.
 *
 * Don't use this as a coding example.  I was trying to find a
 * routing problem and this code sort-of popped out after 48 hours
 * without sleep.  I was amazed it ever compiled, much less ran.
 *
 * I stole the idea for this program from Steve Deering.  Since
 * the first release, I've learned that had I attended the right
 * IETF working group meetings, I also could have stolen it from Guy
 * Almes or Matt Mathis.  I don't know (or care) who came up with
 * the idea first.  I envy the originators' perspicacity and I'm
 * glad they didn't keep the idea a secret.
 *
 * Tim Seaver, Ken Adelman and C. Philip Wood provided bug fixes and/or
 * enhancements to the original distribution.
 *
 * I've hacked up a round-trip-route version of this that works by
 * sending a loose-source-routed udp datagram through the destination
 * back to yourself.  Unfortunately, SO many gateways botch source
 * routing, the thing is almost worthless.  Maybe one day...
 *
 *  -- Van Jacobson (van@helios.ee.lbl.gov)
 *     Tue Dec 20 03:50:13 PST 1988
 *
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */


/*
 * PSC Changes Copyright (c) 1992 Pittsburgh Supercomputing Center.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the modifications to this
 * software were developed by the Pittsburgh Supercomputing Center.
 * The name of the Center may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE
 * IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. 
 *
 */

/* DNS lookup portions of this code also subject to the following:  */
/*
 * Use the domain system to resolve a name.
 *
 * Copyright (C) 1988,1990,1992 Dan Nydick, Carnegie-Mellon University
 * Anyone may use this code for non-commercial purposes as long
 * as my name and copyright remain attached.
 */

/*
 *  1/7/92:  Modified to include a new option for MTU discovery.
 *           This option will send out packets with the don't
 *           fragment bit set in order to determine the MTU
 *           of the path being traceroute'd.  Only decreases
 *           in MTU will be detected, and the MTU will initially
 *           be set to the interface MTU which is used for 
 *           routing.  In the event of an MTU decrease in the 
 *           path, the output will include a message of the 
 *           form MTU=#### with the new MTU for the latest 
 *           hop.  This option is invoked with "-M".
 *                            Jamshid Mahdavi,   PSC.
 */

/*
 * 4/12/93:  Modified to include new option (-Q) that will report on
 *           percentage packet loss* in addition to the usual
 *           min/avg/max delay.  When -Q is invoked, delay per packet
 *           reporting is turned off.  Also ^C aborts further packets
 *           sent on that ttl and goes on to the next -- two
 *           consecutive ^Cs will terminate the traceroute entirely.
 *           [modified code rom Matt Mathis' uping.c code] 
 *                          Jon Boone,  PSC.
 */

/*
 * 4/23/93:  Added support for a "-a" switch which will support
 *           automatic cutoff after a certain number of dropped
 *           packets in a row
 *                          Jon Boone, PSC.
 */

/*
 * 10/21/93: (JM) Fixed SGI version, changed the packet sizing scheme
 *           to a saner system based on total packet size (which
 *           also fixed several bugs in the old code w.r.t. 
 *           size of packets).  Added fast timers.  Added network 
 *           owner lookup.  Plan to add AS path lookup eventually...
 */

#include <stdio.h>
#include <errno.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>

#include <sys/param.h>

#ifdef	vms
#include <sys/types.h>		/* caddr_t, u_short, etc. for socket.h */
#define exit(x) sys$exit(1-x) 	/* in VMS a 1 means ok, and 0 is an error... */
#define index(x,y) strchr(x,y)  /* VAX C run time library lacks index() */
#endif	/* vms */

#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netdb.h>
#include <ctype.h>
#include <math.h>

#define	MAXPACKET	65535	/* max ip packet size */
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	256
#endif

#define SIZEOFstructip sizeof(struct ip)

#ifndef FD_SET
#define NFDBITS         (8*sizeof(fd_set))
#define FD_SETSIZE      NFDBITS
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

#define Fprintf (void)fprintf
#define Sprintf (void)sprintf
#define Printf (void)printf

#define NOERR(val,msg) {if (((int)(val)) < 0) {perror(msg);exit(1);}}

/*  For some reason, IP_HDRINCL and LSRR don't interact well on SGI;
 *  so turn it off:  */
#ifdef __sgi
#undef IP_HDRINCL
#endif

extern	int errno;
extern  char *inet_ntoa();
extern  u_long inet_addr();

void halt();   /* signal handler */

/*
 * format of a (udp) probe packet.
 */
struct opacket {
	struct ip ip;
	struct udphdr udp;
	u_char seq;		/* sequence number of this packet */
	u_char ttl;		/* ttl packet left with */
	struct timeval tv;	/* time packet left */
};

u_char	packet[512];		/* last inbound (icmp) packet */
struct opacket	*outpacket;	/* last output (udp) packet */
char *inetname();
u_char  optlist[MAX_IPOPTLEN];  /* IP options list  */
int _optlen;
struct icmp *icp;               /* Pointer to ICMP header in packet */


int s;				/* receive (icmp) socket file descriptor */
int sndsock;			/* send (udp) socket file descriptor */
struct timezone tz;		/* leftover */

struct sockaddr whereto;	/* Who to try to reach */
int datalen;			/* How much data */

char *source = 0;
char *hostname;
char hnamebuf[MAXHOSTNAMELEN];

int nprobes = 3;
int max_ttl = 30;
u_short ident;
u_short port = 32768+666;	/* start udp dest port # for probe packets */

int options;			/* socket options */
int verbose;
int mtudisc=0;                  /* do MTU discovery in path */
float waittime = 3.0;		/* time to wait for response (in seconds) */
int nflag;			/* print addresses numerically */

int haltf=0;                    /* signal happened */
int ppdelay=1;                  /* we normally want per-packet delay */
int pploss=0;                   /* we normally don't want packet loss */
int lost;                       /* how many packets did we not get back */ 
double throughput;              /* percentage packets not lost */
int consecutive=0;              /* the number of consecutive lost packets */
int automagic=0;                /* automatically quit after 10 lost packets? */

int utimers=0;                  /* Print timings in microseconds */
int dns_owner_lookup=0;         /* Look up owner email in DNS */
int as_lookup=0;                /* Look up AS path in routing registries */

char usage[] = "Usage: traceroute [-danruvAMOQ] [-w wait] [-m max_ttl] [-p port#] [-q nqueries] [-t tos] [-s src_addr] [-g router] host [data size]\n\
      -a: Abort after 10 consecutive drops\n\
      -d: Socket level debugging\n\
      -g: Use this gateway as an intermediate hop (uses LSRR)\n\
      -m: Set maximum TTL (default 30)\n\
      -n: Report IP addresses only (not hostnames)\n\
      -p: Use an alternate UDP port\n\
      -q: Set the number of queries at each TTL (default 3)\n\
      -r: Set Dont Route option\n\
      -s: Set your source address\n\
      -t: Set the IP TOS field (default 0)\n\
      -u: Use microsecond timestamps\n\
      -v: Verbose\n\
      -w: Set timeout for replies (default 5 sec)\n\
      -A: Report AS# at each hop (from GRR) (Not implemented)\n\
      -M: Do RFC1191 path MTU discovery\n\
      -S: Set the starting TTL (default=1)\n\
      -O: Report owner at each hop (from DNS)\n\
      -Q: Report delay statistics at each hop (min/avg+-stddev/max) (ms)\n";

float deltaT();

main(argc, argv)
        int argc;
	char *argv[];
{
	struct sockaddr_in from;
	char **av = argv;
	struct sockaddr_in *to = (struct sockaddr_in *) &whereto;
	int on = 1;
	int alloc_len;
	struct protoent *pe;
	int ttl, probe, i;
	int seq = 0;
	int tos = 0;
	struct hostent *hp;
	int lsrr = 0;
	u_long gw;
	u_char *oix;
	int new_mtu = 0;
	int min_ttl=1;

	oix = optlist;
	bzero(optlist, sizeof(optlist));

#ifdef vms
	if (argc < 3) fixargs(&argc,argv,av);	
#endif
	argc--, av++;
	while (argc && *av[0] == '-')  {
		while (*++av[0])
			switch (*av[0]) {
			case 'a':
			        automagic = 1;
				break;
			case 'A':
				as_lookup = 1;
				break;
			case 'd':
				options |= SO_DEBUG;
				break;
			case 'g':
				argc--, av++;
				if ((lsrr+1) >= ((MAX_IPOPTLEN-IPOPT_MINOFF)/sizeof(u_long))) {
				  Fprintf(stderr,"No more than %d gateway\n",
					  ((MAX_IPOPTLEN-IPOPT_MINOFF)/sizeof(u_long))-1);
				  exit(1);
				}
				if (lsrr == 0) {
				  *oix++ = IPOPT_LSRR;
				  *oix++;	/* Fill in total length later */
				  *oix++ = IPOPT_MINOFF; /* Pointer to LSRR addresses */
				}
				lsrr++;
				if (isdigit(*av[0])) {
				  gw = inet_addr(*av);
				  if (gw) {
				    bcopy(&gw, oix, sizeof(u_long));
				  } else {
				    Fprintf(stderr, "Unknown host %s\n",av[0]);
				    exit(1);
				  }
				} else {
				  hp = gethostbyname(av[0]);
				  if (hp) {
				    bcopy(hp->h_addr, oix, sizeof(u_long));
				  } else {
				    Fprintf(stderr, "Unknown host %s\n",av[0]);
				    exit(1);
				  }
				}
				oix += sizeof(u_long);
				goto nextarg;
			case 'm':
				argc--, av++;
				max_ttl = atoi(av[0]);
				goto nextarg;
			case 'n':
				nflag++;
				break;
			case 'O':
				dns_owner_lookup = 1;
				break;
			case 'p':
				argc--, av++;
				port = atoi(av[0]);
				if (port < 1) {
					Fprintf(stderr, "port must be >0\n");
					exit(1);
				}
				goto nextarg;
			case 'q':
				argc--, av++;
				nprobes = atoi(av[0]);
				if (nprobes < 1) {
					Fprintf(stderr, "nprobes must be >0\n");
					exit(1);
				}
				goto nextarg;
			case 'r':
				options |= SO_DONTROUTE;
				break;
			case 's':
				/*
				 * set the ip source address of the outbound
				 * probe (e.g., on a multi-homed host).
				 */
				argc--, av++;
				source = av[0];
				goto nextarg;
			case 'S':
				argc--, av++;
				min_ttl = atoi(av[0]);
				goto nextarg;
			case 't':
				argc--, av++;
				tos = atoi(av[0]);
				if (tos < 0 || tos > 255) {
					Fprintf(stderr, "tos must be 0 to 255\n");
					exit(1);
				}
				goto nextarg;
			case 'u':
				utimers = 1;
				break;
			case 'v':
				verbose++;
				break;
			case 'w':
				argc--, av++;
				waittime = atof(av[0]);
				if (waittime <= .01) {
					Fprintf(stderr, "wait must be >10 msec\n");
					exit(1);
				}
				goto nextarg;
 		        case 'M':
				mtudisc++;
				break;
                        case 'Q':
				pploss = 1;
				ppdelay = 0;
				break;
			default:
				Printf(usage);
				exit(1);
			}
	nextarg:
		argc--, av++;
	}
	if (argc < 1)  {
		Printf(usage);
		exit(1);
	}
#ifndef vms
	setlinebuf (stdout);
#endif

	if (max_ttl < min_ttl) {
	  Fprintf(stderr, "max ttl must be >= min_ttl.\n");
	  exit(1);
	}

	(void) bzero((char *)&whereto, sizeof(struct sockaddr));
	to->sin_family = AF_INET;
	to->sin_addr.s_addr = inet_addr(av[0]);
	if (to->sin_addr.s_addr != -1) {
		(void) strcpy(hnamebuf, av[0]);
		hostname = hnamebuf;
	} else {
		hp = gethostbyname(av[0]);
		if (hp) {
			to->sin_family = hp->h_addrtype;
			bcopy(hp->h_addr, (caddr_t)&to->sin_addr, hp->h_length);
			hostname = hp->h_name;
		} else {
			Printf("%s: unknown host %s\n", argv[0], av[0]);
			exit(1);
		}
	}

	if (argc >= 2)
		datalen = atoi(av[1]);
	if (datalen < 0 || datalen >= MAXPACKET) {
		Fprintf(stderr, "traceroute: packet size must be 0 <= s < %ld\n",
			MAXPACKET - sizeof(struct opacket));
		exit(1);
	}
	if (mtudisc) 
	  /*  Ignore data length as set.  Set it to a large value to 
	      start things off...  */
	  datalen=32000;

	if (datalen<sizeof(struct opacket) + MAX_IPOPTLEN) {
	  alloc_len = sizeof(struct opacket) + MAX_IPOPTLEN;
	} else {
	  alloc_len = datalen;
	}

	outpacket = (struct opacket *)malloc((unsigned)alloc_len);

	if (! outpacket) {
		perror("traceroute: malloc");
		exit(1);
	}
	(void) bzero((char *)outpacket, alloc_len);
	outpacket->ip.ip_dst = to->sin_addr;
	outpacket->ip.ip_tos = tos;

	ident = (getpid() & 0xffff) | 0x8000;

	if ((pe = getprotobyname("icmp")) == NULL) {
		Fprintf(stderr, "icmp: unknown protocol\n");
		exit(10);
	}
	if ((s = socket(AF_INET, SOCK_RAW, pe->p_proto)) < 0) {
		perror("traceroute: icmp socket");
		exit(5);
	}
	if (options & SO_DEBUG)
		(void) setsockopt(s, SOL_SOCKET, SO_DEBUG,
				  (char *)&on, sizeof(on));
	if (options & SO_DONTROUTE)
		(void) setsockopt(s, SOL_SOCKET, SO_DONTROUTE,
				  (char *)&on, sizeof(on));

	if ((sndsock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
		perror("traceroute: raw socket");
		exit(5);
	}

#ifndef vms
	/*  ^C punts you to the next hop.  Twice will exit.  */
	NOERR(signal(SIGINT,halt),"signal SIGINT");   
#endif
	
	if (lsrr > 0) {
	  lsrr++;
	  optlist[IPOPT_OLEN]=IPOPT_MINOFF-1+(lsrr*sizeof(u_long));
	  bcopy((caddr_t)&to->sin_addr, oix, sizeof(u_long));
	  oix += sizeof(u_long);
	  while ((oix - optlist)&3) oix++;		/* Pad to an even boundry */
	  _optlen = (oix - optlist);

	  if ((pe = getprotobyname("ip")) == NULL) {
	    perror("traceroute: unknown protocol ip\n");
	    exit(10);
	  }
	  if ((setsockopt(sndsock, pe->p_proto, IP_OPTIONS, optlist, oix-optlist)) < 0) {
	    perror("traceroute: lsrr options");
	    exit(5);
	  }
	}

	if (datalen < sizeof (struct opacket) + _optlen) {
	  /*  The chosen size is too small to fit everything...
	      make it bigger:  */
	  datalen = sizeof (struct opacket) + _optlen;
      }


#ifdef SO_SNDBUF
	if (setsockopt(sndsock, SOL_SOCKET, SO_SNDBUF, (char *)&datalen,
		       sizeof(datalen)) < 0) {
		perror("traceroute: SO_SNDBUF");
		exit(6);
	}
#endif SO_SNDBUF
#ifdef IP_HDRINCL
	if (setsockopt(sndsock, IPPROTO_IP, IP_HDRINCL, (char *)&on,
		       sizeof(on)) < 0) {
		perror("traceroute: IP_HDRINCL");
		exit(6);
	}
#endif IP_HDRINCL
	if (options & SO_DEBUG)
		(void) setsockopt(sndsock, SOL_SOCKET, SO_DEBUG,
				  (char *)&on, sizeof(on));
	if (options & SO_DONTROUTE)
		(void) setsockopt(sndsock, SOL_SOCKET, SO_DONTROUTE,
				  (char *)&on, sizeof(on));

	if (source) {
		(void) bzero((char *)&from, sizeof(struct sockaddr));
		from.sin_family = AF_INET;
		from.sin_addr.s_addr = inet_addr(source);
		if (from.sin_addr.s_addr == -1) {
			Printf("traceroute: unknown host %s\n", source);
			exit(1);
		}
		outpacket->ip.ip_src = from.sin_addr;
#ifndef IP_HDRINCL
		if (bind(sndsock, (struct sockaddr *)&from, sizeof(from)) < 0) {
			perror ("traceroute: bind:");
			exit (1);
		}
#endif IP_HDRINCL
	}

	Fprintf(stderr, "traceroute to %s (%s)", hostname,
		inet_ntoa(to->sin_addr));
	if (source)
		Fprintf(stderr, " from %s", source);
	Fprintf(stderr, ", %d hops max, %d byte packets\n", max_ttl, datalen);
	(void) fflush(stderr);

	for (ttl = min_ttl; ttl <= max_ttl; ++ttl) {
		u_long lastaddr = 0;
		int got_there = 0;
		int unreachable = 0;
		float min, max, sum = 0, sumsq = 0;
		int mtu=0;   /*  A temp variable; new_mtu and datalen have
				 the real numbers  */

		Printf("%2d ", ttl);
		(void) fflush(stdout);
		if (new_mtu != 0) {
		  Printf("MTU=%d ",new_mtu);
		  new_mtu=0;
		}
		lost = 0;
		throughput = (double) 0.0;
		consecutive = 0;
		for (probe = 0; probe < nprobes; ++probe) {
			int cc;
			struct timeval tv;
			struct ip *ip;

			(void) gettimeofday(&tv, &tz);
			send_probe(++seq, ttl);
			while (cc = wait_for_reply(s, &from)) {
				if ((i = packet_ok(packet, cc, &from, seq))) {
					float dt = deltaT(&tv);
					if (sum == 0) {
					    sum = min = max = dt;
					    sumsq = dt*dt;
				       } else {
					    if (dt < min) min = dt;
					    if (dt > max) max = dt;
					    sum += dt;
					    sumsq += dt*dt;
				       }

					if (from.sin_addr.s_addr != lastaddr) {
						print(packet, cc, &from);
						lastaddr = from.sin_addr.s_addr;
					}
					if (ppdelay) {
					  if (utimers)
					    Printf("  %3.3f ms", dt);
					  else
					    Printf("  %d ms", (int) (dt+0.5));
					  (void) fflush(stdout);
					}

					switch(i - 1) {
					case ICMP_UNREACH_PORT:
#ifndef ARCHAIC
						ip = (struct ip *)packet;
						if (ip->ip_ttl <= 1)
							Printf(" !");
#endif ARCHAIC
						++got_there;
						break;
					case ICMP_UNREACH_NET:
						++unreachable;
						Printf(" !N");
						break;
					case ICMP_UNREACH_HOST:
						++unreachable;
						Printf(" !H");
						break;
					case ICMP_UNREACH_PROTOCOL:
						++got_there;
						Printf(" !P");
						break;
					case ICMP_UNREACH_NEEDFRAG:
						if (mtudisc) {
						  mtu = (ntohl(icp->icmp_void) & 0xffff);
						  if (mtu >= datalen) {
						    /*  This should never
							happen.  There is
							a serious bug 
							somewhere... */
						    Printf (" !M>");
						  } 
						  else if (mtu == 0) {
						    Printf (" !M");
						    new_mtu = reduce_mtu(datalen);
						  }
						  else {
						    new_mtu = mtu;
						    Printf (" !M=%d", new_mtu);
						  }
						  break;
						}
						else {
						  ++unreachable;
						  Printf(" !F");
						  break;
						}
					case ICMP_UNREACH_SRCFAIL:
						++unreachable;
						Printf(" !S");
						break;
					}
					consecutive = 0; /* got a packet back! */
					break;
				   }
			}
			if (cc == 0) {
			     Printf(" *");
			     (void) fflush(stdout);

			     lost++;
			     consecutive++;

			     /*  Reset the ^C action from exit to skip TTL  */
#ifndef vms
       			     if (haltf==0 && lost==1)
			       NOERR(signal(SIGINT,halt), "signal SIGINT");
#endif /* vms */
			     /* we've missed at least one packet, so */
			     /* let's check for the signal to go to */
			     /* the next ttl */
			     if (haltf > 0) {
				  haltf = 0;
				  consecutive = 0;
				  break;
			     }
			}

			if (automagic && (consecutive > 9))
			     break;
		   }

		if (pploss) {
		     if (lost < probe) {
			  throughput = 
			       ( 100.0 - ( ( lost * 100.0 ) / probe ));
			  printf("  (%1.1f ms/%1.1f ms(+-%1.1f ms)/%1.1f ms)", 
				 min, (sum / (probe - lost)), 
				 (float)sqrt((double)sumsq)/(probe-lost), max);
			  printf(" %d/%d (%#3.2f%%)", (probe - lost),
				 probe, throughput);
			  (void) fflush(stdout);
		     }
		}
		putchar('\n');
		if (got_there || unreachable > nprobes - 1)
			exit(0);
		if (new_mtu != 0) {
		  ttl--;  /*  Redo the same TTL  */
		  datalen = new_mtu;  /*  Set the new data length  */
		}
	  }
}

wait_for_reply(sock, from)
	int sock;
	struct sockaddr_in *from;
{
	fd_set fds;
	struct timeval wait;
	int cc = 0;
	int fromlen = sizeof (*from);

	FD_ZERO(&fds);
	FD_SET(sock, &fds);
	wait.tv_sec = (int) waittime; 
	wait.tv_usec = ((int) (waittime*1000000.0))%1000000;

	if (select(sock+1, &fds, (fd_set *)0, (fd_set *)0, &wait) > 0)
		cc=recvfrom(s, (char *)packet, sizeof(packet), 0,
			    (struct sockaddr *)from, &fromlen);

	return(cc);
}


send_probe(seq, ttl)
{
	struct opacket *op = outpacket;
	struct ip *ip = &op->ip;
	struct udphdr *up = &op->udp;
	int i;

      retry:
	if (mtudisc) {
	  ip->ip_off = (short) IP_DF;
	}
	else
	  ip->ip_off = 0;
	ip->ip_p = IPPROTO_UDP;
	ip->ip_len = datalen-_optlen;   /*  The OS inserts options  */
	ip->ip_ttl = ttl;

	up->uh_sport = htons(ident);
	up->uh_dport = htons(port+seq);
	up->uh_ulen = htons((u_short)(datalen - sizeof(struct ip) - _optlen));
	up->uh_sum = 0;

	op->seq = seq;
	op->ttl = ttl;
	(void) gettimeofday(&op->tv, &tz);

	i = sendto(sndsock, (char *)outpacket, datalen - _optlen, 0, &whereto,
		   sizeof(struct sockaddr));

	if (i < 0 || i != datalen - _optlen)  {
		if (i<0) {
		    if (errno == EMSGSIZE) {
		        datalen = reduce_mtu(datalen);
			Printf (" MTU=%d", datalen);
			goto retry;
		    }
		    else
			perror("sendto");
		}
		Printf("traceroute: wrote %s %d chars, ret=%d\n", hostname,
			datalen, i);
		(void) fflush(stdout);
	}
}


float deltaT(tp)
	struct timeval *tp;
{
	struct timeval tv;

	(void) gettimeofday(&tv, &tz);
	tvsub(&tv, tp);
	return (tv.tv_sec*1000.0 + ((float)tv.tv_usec)/1000.0);
}


/*
 * Convert an ICMP "type" field to a printable string.
 */
char *
pr_type(t)
	u_char t;
{
	static char *ttab[] = {
	"Echo Reply",	"ICMP 1",	"ICMP 2",	"Dest Unreachable",
	"Source Quench", "Redirect",	"ICMP 6",	"ICMP 7",
	"Echo",		"ICMP 9",	"ICMP 10",	"Time Exceeded",
	"Param Problem", "Timestamp",	"Timestamp Reply", "Info Request",
	"Info Reply"
	};

	if(t > 16)
		return("OUT-OF-RANGE");

	return(ttab[t]);
}


packet_ok(buf, cc, from, seq)
	u_char *buf;
	int cc;
	struct sockaddr_in *from;
	int seq;
{
	u_char type, code;
	int hlen, mtu;
#ifndef ARCHAIC
	struct ip *ip;

	ip = (struct ip *) buf;
	hlen = ip->ip_hl << 2;
	if (cc < hlen + ICMP_MINLEN) {
		if (verbose)
			Printf("packet too short (%d bytes) from %s\n", cc,
				inet_ntoa(from->sin_addr));
		return (0);
	}
	cc -= hlen;
	icp = (struct icmp *)(buf + hlen);
#else
	icp = (struct icmp *)buf;
#endif ARCHAIC
	type = icp->icmp_type; code = icp->icmp_code;
	if ((type == ICMP_TIMXCEED && code == ICMP_TIMXCEED_INTRANS) ||
	    type == ICMP_UNREACH) {
		struct ip *hip;
		struct udphdr *up;
		      
		hip = &icp->icmp_ip;
		hlen = hip->ip_hl << 2;
		up = (struct udphdr *)((u_char *)hip + hlen);
		if (hlen + 12 <= cc && hip->ip_p == IPPROTO_UDP &&
		    up->uh_sport == htons(ident) &&
		    up->uh_dport == htons(port+seq))
			return (type == ICMP_TIMXCEED? -1 : code+1);
	}
#ifndef ARCHAIC
	if (verbose) {
		int i;
		u_long *lp = (u_long *)&icp->icmp_ip;

		Printf("\n%d bytes from %s to %s", cc,
			inet_ntoa(from->sin_addr), inet_ntoa(ip->ip_dst));
		Printf(": icmp type %d (%s) code %d\n", type, pr_type(type),
		       icp->icmp_code);
		for (i = 4; i < cc ; i += sizeof(long))
			Printf("%2d: x%8.8lx\n", i, *lp++);
	}
#endif ARCHAIC
	return(0);
}

int reduce_mtu(value)
int value;
{
  /*  The following heuristic taken from RFC1191  */
  static int mtuvals[11]={32000, 17914, 8166, 4352, 2002, 1492, 1006, 508, 296, 68, -1};
  int i=0;

  while (value <= mtuvals[i]) i++;
  if (mtuvals[i] > 0) {
    value = mtuvals[i];
  }
  else {
    Printf (" No valid MTU!!!\n");
    exit (1);
  }
  return (value);
}

char *lookup_owner();
char *doresolve ();

print(buf, cc, from)
	u_char *buf;
	int cc;
	struct sockaddr_in *from;
{
	struct ip *ip;
	int hlen;

	ip = (struct ip *) buf;
	hlen = ip->ip_hl << 2;
	cc -= hlen;

	if (nflag)
		Printf(" %s", inet_ntoa(from->sin_addr));
	else
		Printf(" %s (%s)", inetname(from->sin_addr),
		       inet_ntoa(from->sin_addr));

	if (dns_owner_lookup)
	  Printf(" [%s]", lookup_owner(from->sin_addr));
	
	if (verbose)
	  Printf (" %d bytes to %s", cc, inet_ntoa (ip->ip_dst));
      }


#ifdef notyet
/*
 * Checksum routine for Internet Protocol family headers (C Version)
 */
in_cksum(addr, len)
     u_short *addr;
     int len;
{
  register int nleft = len;
  register u_short *w = addr;
  register u_short answer;
  register int sum = 0;

  /*
   *  Our algorithm is simple, using a 32 bit accumulator (sum),
   *  we add sequential 16 bit words to it, and at the end, fold
   *  back all the carry bits from the top 16 bits into the lower
   *  16 bits.
   */
  while (nleft > 1)  {
    sum += *w++;
    nleft -= 2;
  }

  /* mop up an odd byte, if necessary */
  if (nleft == 1)
    sum += *(u_char *)w;

  /*
   * add back carry outs from top 16 bits to low 16 bits
   */
  sum = (sum >> 16) + (sum & 0xffff); /* add hi 16 to low 16 */
  sum += (sum >> 16);		/* add carry */
  answer = ~sum;		/* truncate to 16 bits */
  return (answer);
}
#endif notyet

/*
 * Subtract 2 timeval structs:  out = out - in.
 * Out is assumed to be >= in.
 */
tvsub(out, in)
     register struct timeval *out, *in;
{
  if ((out->tv_usec -= in->tv_usec) < 0)   {
    out->tv_sec--;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}


/*
 * Construct an Internet address representation.
 * If the nflag has been supplied, give 
 * numeric value, otherwise try for symbolic name.
 */
char *
inetname(in)
     struct in_addr in;
{
  register char *cp;
  static char line[50];
  struct hostent *hp;
  static char domain[MAXHOSTNAMELEN + 1];
  static int first = 1;

  if (first && !nflag) {
    first = 0;
    if (gethostname(domain, MAXHOSTNAMELEN) == 0 &&
	(cp = index(domain, '.')))
      (void) strcpy(domain, cp + 1);
    else
      domain[0] = 0;
  }
  cp = 0;
  if (!nflag && in.s_addr != INADDR_ANY) {
    hp = gethostbyaddr((char *)&in, sizeof (in), AF_INET);
    if (hp) {
      if ((cp = index(hp->h_name, '.')) &&
	  !strcmp(cp + 1, domain))
	*cp = 0;
      cp = hp->h_name;
    }
  }
  if (cp)
    (void) strcpy(line, cp);
  else {
    in.s_addr = ntohl(in.s_addr);
#define C(x)	((x) & 0xff)
    Sprintf(line, "%lu.%lu.%lu.%lu", C(in.s_addr >> 24),
	    C(in.s_addr >> 16), C(in.s_addr >> 8), C(in.s_addr));
  }
  return (line);
}

void halt()
{
  haltf++;
#ifndef vms
  NOERR(signal(SIGINT,0), "signal SIGINT,0");
#endif
}


/*
 *  Lookup owner of the net in DNS.
 */


char *lookup_owner(in)
     struct in_addr in;
{
  char dns_query[100];
  char *owner, *dot_ptr;
  unsigned char *addr_ptr;

  addr_ptr = (unsigned char *) (&in.s_addr);
  
  if (addr_ptr[0] >= 192) {
    sprintf (dns_query, "%d.%d.%d.in-addr.arpa", addr_ptr[2], addr_ptr[1], addr_ptr[0]);
  } else if (addr_ptr[0] >= 128) {
    sprintf (dns_query, "%d.%d.in-addr.arpa", addr_ptr[1], addr_ptr[0]);
  } else {
    sprintf (dns_query, "%d.in-addr.arpa", addr_ptr[0]);
  }
  
  owner = doresolve (dns_query);
  
  /*  reformat slightly  */
  if (owner != NULL) {
    if ( strchr (owner, '@') == NULL ) {
      dot_ptr = strchr (owner, '.');
      if (dot_ptr != NULL) 
	*dot_ptr = '@';
    }
    
    if (strlen(owner) > 0) {
      dot_ptr=owner + strlen (owner) - 1;
      while (*dot_ptr == ' ' || *dot_ptr == '.' ) {
	*dot_ptr = 0;
        dot_ptr--;
      }
    }
  }

  return (owner);

}

#include <arpa/nameser.h>

#ifndef vms
#include <resolv.h>
#else
#include "multinet_root:[multinet.include]resolv.h"
#endif

short getshort(ptr)
char *ptr;
{
    union {
	short retval;
	char ch[2];
    } foo;

    foo.ch[0] = (*ptr & 0xff);
    foo.ch[1] = (*(ptr+1) & 0xff);

    return (foo.retval);
}

char *doresolve (name)
char *name;
{
  int query=QUERY;
  int qtype=T_SOA;
  int qclass=C_IN;
  unsigned char buf[256];
  char *ans;
  int blen, alen, got;
  int anssiz, i;
  short shrt;
  HEADER *h;
  char *contact_ptr=0;
  int ptr;

  anssiz = 512;
  ans = malloc(anssiz);
  if (!ans) {
    return(0);
  }

  blen = res_mkquery(query,name,qclass,qtype,NULL,0,NULL,buf,sizeof(buf));
  if (blen < 0) {
    return (0);
  }

  alen = res_send(buf,blen,ans,anssiz);
  if (alen == -1) {
    return (0);
  }
 
  if (alen < 12) {
    return (0);
  }

  h = (HEADER *)ans;
  
  h->id = ntohs(h->id);
  h->qdcount = ntohs(h->qdcount);
  h->ancount = ntohs(h->ancount);
  h->nscount = ntohs(h->nscount);
  h->arcount = ntohs(h->arcount);

  ptr = 12;	/* point at first question field */
  for (i=0; i<h->qdcount && ptr<alen; i++) {
    ptr=doqd(ans,ptr);
  }

  for (i=0; i<h->ancount && ptr<alen; i++) {
    ptr=dorr(ans,ptr,&contact_ptr);
  }

  return (contact_ptr);
}

doqd(ans,off)
unsigned char *ans;
int off;
{
    char name[256];

    name[0]=0;
    off = doname(ans,off,name);
    off = dotype(ans,off);
    off = doclass(ans,off);
    return (off);
}

dorr(ans,off,contact_ptr)
unsigned char *ans;
int off;
char **contact_ptr;
{
    int class, typ;
    char name[256];

    name[0]=0;
    off = doname(ans,off,name);
    typ = ntohs(getshort(ans+off));
    off = dotype(ans,off);
    class = ntohs(getshort(ans+off));
    off = doclass(ans,off);
    off = dottl(ans,off);
    off = dordata(ans,off,class,typ,name,contact_ptr);
    return(off);
}

doname(ans,off,name)
int off;
unsigned char *ans;
char *name;
{
    int newoff, i;
    char tmp[50];

	/* redirect? */
    if ((*(ans+off) & 0xc0) == 0xc0) {
	newoff = getshort(ans+off);
	newoff = 0x3fff & ntohs(newoff);
	doname(ans,newoff,name);
	return (off+2);
    }
	/* end of string */
    if (*(ans+off) == 0) {
	strcat(name," ");
	return(off+1);
    }
	/* token */
    for (i=1; i<=*(ans+off); i++)
	tmp[i-1]=ans[off+i];
    tmp[i-1] = '.';
    tmp[i]=0;
    strcat(name,tmp);
    return (doname(ans,off+1+(*(ans+off)), name));
}

dotype(ans,off)
int off;
unsigned char *ans;
{
    return(off+2);
}

doclass(ans,off)
int off;
unsigned char *ans;
{
    return(off+2);
}

dottl(ans,off)
int off;
unsigned char *ans;
{
    return(off+4);
}

dordata(ans,off,class,typ,fname,contact_ptr)
unsigned char *ans;
int off,class,typ;
char *fname;
char **contact_ptr;
{
    int len = ntohs(getshort(ans+off));
    int retval = off+len+2;
    int i,j;
    char name[256];

    off += 2;
    switch (typ) {
	case T_SOA:
	    name[0]=0;
	    off = doname(ans,off,name);
	    name[0]=0;
	    off = doname(ans,off,name);
	    *contact_ptr=name;
	    return (0);
	default:
	    return (0);
    }
}

/*
	The VMS command line interface, DCL, uppercases all unquoted input.
	By default this program can be executed via
		$ mc location:traceroute options

	However, since this program needs case sensitivity for the various 
	options to work, it is defined as an alias ``symbol'' with open 
	quotes:
		$ traceroute == "location:traceroute """

	Unfortunately, this has the side effects of making argc=2, where 
	argv[0]	is the image location and name, and argv[1] is the entire 
	option line.  Thus for VMS we need to break the options up...
*/

#ifdef vms		
			
fixargs(a,b,c)
int *a;			/* argc */
char **b;		/* argv */
char **c;		/* av */
{	
	char *ptr, *space;
	/* Set the image name */
	c[0] = b[0];

	/* Initialize pointers */
	*a = 1;
	ptr = b[1];
	if (*ptr == ' ') ptr++;		/* eliminate first space */

	/* Delineate all strings ending with space */
	while ((space = strchr(ptr,' ')) != 0) {
	   *space = '\0';
	   c[(*a)++] = ptr;
	   ptr = space + 1;
	}
	
	/* Transfer last one - ending with null */
	c[*a] = ptr;

	/* Update argc */
	(*a)++;
}

#endif /* vms */


